/*
chil.h

Header for chil.c

*/

/* First the preprocessor definitions -
Many were taken from various header files used by Mark Blackman's rad93.c
*/

#define CHIL_LAT  51.1445
#define CHIL_LONG -1.4370

/* System IDs: these values are contained in byte 29 of the IR in some recent data */
#define UNKNOWN_SYSTEM 0
#define CAMRA 1
#define RABELAIS 2
#define GALILEO 3
#define SINGAPORE 4

/* System frequencies */
#define CAMRA_FREQ              3.075 /* tx freq, Ghz */		  
#define RABELAIS_FREQ            34.960 /* number from official documents */
#define GALILEO_FREQ  94.00      /* approximately... */
#define SINGAPORE_FREQ 3.0   /* approximately */

/* Parameter IDs */
#define CAMRA_ZH	1
#define CAMRA_ZDR	2
#define CAMRA_LDR	3
#define CAMRA_V	4
#define CAMRA_WIDTH	5
#define CAMRA_KDP	6
#define CAMRA_PHIDP	7
#define CAMRA_VH	8
#define CAMRA_VV	9
#define CAMRA_V2	10
#define RABELAIS_ZH	30
#define RABELAIS_ZDR	31
#define RABELAIS_PHIDP	32
#define RABELAIS_V	33
#define RABELAIS_VV	34
#define SINGAPORE_ZH	40
#define SINGAPORE_LDR	41
#define SINGAPORE_V	42
#define SINGAPORE_WIDTH	43
#define GALILEO_ZH	50
#define GALILEO_V	51

/* scan types */
#define RHI_93	1
#define PPI_93	2
#define SP_93	3
#define FIXED_93	4
#define CAL_93	5
#define TS_93	6
#define UF_CAL 0
#define UF_PPI 1
#define UF_COPLANE 2
#define UF_RHI 3
#define UF_VERT 4
#define UF_DWELL 5
#define UF_MANUAL 6
#define UF_IDLE 7

/* New format bitwise parameter IDs and corresponding masks*/
#define UNKNOWN 0
#define ZH 1
#define ZDR 2
#define LDR 3
#define PHIDP 4
#define V1 5
#define V2 6
#define WIDTH_H 7
#define WIDTH_V 8
/* #define WIDTH 7 */
#define V_STD 15

#define ZH_MASK (1L<<0)
#define ZDR_MASK (1L<<1)
#define LDR_MASK (1L<<2)
#define PHIDP_MASK (1L<<3)
#define V1_MASK (1L<<4)
#define V2_MASK (1L<<5)
#define WIDTH_MASK (1L<<6)
#define OVERFLOW_MASK (1L<<7)

#define UNKNOWN_NAME "Unknown"
#define ZH_NAME   "Zh"
#define ZDR_NAME   "Zdr"
#define LDR_NAME   "Ldr"
#define PHIDP_NAME  "phi_dp"
#define VEL1_NAME   "v1"
#define VEL2_NAME   "v2"
#define WIDTH_NAME "width"      
#define WIDTH_H_NAME "width_h"      
#define WIDTH_V_NAME "width_v"      
#define ZH1_NAME   "Zh"
#define ZH2_NAME   "Zh2"
#define VEL_NAME   "v"
#define VELSTD_NAME "sigma_v"

#define UNKNOWN_UNITS "-"
#define ZH_UNITS   "dBZ"
#define ZDR_UNITS   "dB"
#define LDR_UNITS   "dB"
#define PHIDP_UNITS  "degrees"
#define VEL1_UNITS   "m/s"
#define VEL2_UNITS   "m/s"
#define WIDTH_UNITS "m/s"      
#define ZH1_UNITS   "dBZ"
#define ZH2_UNITS   "dBZ"
#define VEL_UNITS   "m/s"
#define VELSTD_UNITS "m/s"

/* Scalings */
#define ZH_SCALING_93  0.25
#define ZDR_SCALING_93 0.125
#define LDR_SCALING_93 0.25
#define ZH_SCALING_HP  0.01
#define ZDR_SCALING_HP 0.01
#define PHASE_SCALING_93 0.001
#define MIN_SPACING 37.5
#define MIN_SPACING_GALILEO_NEW 30.0
/* mb. 27/07/95 ....... all according to kev's document of oct. 1994 */
#define END_OF_SCAN 65281
#define END_OF_RASTER 61440
#define END_OF_RASTER_B 61540
#define RASTER_ABORTED 65535

/* for ChilboltonFile structure element `position' */
#define FILE_CLOSED -1
#define START_OF_IR 0
#define START_OF_DATA 1
#define START_OF_RAY 2
#define START_OF_FILE 3
#define UNKNOWN_POSITION 127

/*
for noise subtraction. These values are SYSTEM dependent -
here appropriate for the galileo.
*/
#define MIN_V_STD 0.05
#define MAX_V_STD 2.5

#define BLKSIZE 4092
#define uword unsigned short int
#define word short int
#define uchar unsigned char
#define schar char


#define packvar __attribute__((packed)) /* probably redundant - a Blackman legacy */

#define MAX_FIELDS 15
#define MAX_FIELDNAME_LENGTH 6
#define MAX_FILENAME_LENGTH 255
#define MAX_FILES 8

#include <stdio.h>

/* Now for some useful structures */
typedef struct {
   char n;
   char buffer[MAX_FILES*MAX_FILENAME_LENGTH];
   char *name[MAX_FILES];
   unsigned short int first[MAX_FILES],last[MAX_FILES];
} FileList;

typedef struct {
   FILE 
      *file;
   uchar
      is_raw,   /* flag to indicate that the file is raw - 
                     i.e. contains one raster with the normal byte order reversed */
      summary_exists;   /* indicates that a summary file has been found and read */
   char
      name[MAX_FILENAME_LENGTH],
      default_sum_dir[MAX_FILENAME_LENGTH],
      position;
   uword
      iraster, /* the index of the current raster in the file, starting at 0 */
      iscan; /* index of the current scan */
   /* the following are read from the summary file - undefined if it can't be found */
   uword
      nrasters, /* number of rasters in the file */
      raster[1000], /* array of raster numbers in the order they appear */
      block[1000], /* corresponding block locations of each raster */
      day[1000],
      month[1000],
      year[1000];
   int
      bytes_read; /* bytes of current raster read */
} ChilboltonFile;

typedef struct {
    uchar
      raw[64];   /* raw information record data */
    uchar
      recording_flag packvar,	/* >0, MT on (recording) */
      spacing_raw packvar,
      spacing_ave packvar,
      nfields,
      field_id[MAX_FIELDS] packvar,	/* ID code for each data field, 1 = Zh, 2 =
			 * Zdr, 3 = Ldr, 4 = V, 5 = deltaV, 6 = Kdp,
			 * 7 = Phidp, 8 = Vh, 9 = Vv, 10=Vhv ... */
      field_type[MAX_FIELDS] packvar, /* similar to above but system-independent */
     nscans packvar,		/* no. of scans in raster */
      expt_id packvar,		/* experiment identifier */
      scan_type packvar,		/* scan type,
				 * 1..rhi,2..ppi,3..sp,4..fixed,5..cal
				 * ibration,6..time series */
      sweep_mode packvar,		/* sweep_mode, universal format */
      sens_flag packvar,		/* 1 = 20db of atten., 0..no atttenuation */
      hour packvar,			/* hour of raster start */
      operator[3] packvar,		/* 3 ASCII bytes identifying operater */
      minute packvar,		/* min. of raster start */
      second packvar,		/* second of raster start */
      cmpl_scans packvar,		/* scans in raster completed (first scan = 1) */
      system_id packvar,
      range_normalised packvar,
      time_stamping packvar,
      cloud_data packvar;
   schar
      scan_velocity packvar;	/* scan velocity = deg/sec*50.0 */
//      *field_name[MAX_FIELDNAME_LENGTH];
   uword
      pulse_pairs packvar,
      min_gate packvar,	/* minimum gate, def: 17 */
      max_gate packvar,	/* maximum gate, def: 273 */
      day packvar,		/* day of month scan was taken */
      month packvar,		/* month of scan */
      year packvar,		/* year of scan */
      file packvar;		/* file number, (5...), like old tape numbers */
   word
      threshold packvar,	/* threshold for (displayed ??) Zh in dbZ.
			 * def: -32, 3gig only */
      min_angle packvar,	/* min. angle in raster (arcmin), elevation
			 * for RHI */
      max_angle packvar,	/* max. angle in raster (arcmin), azimuth for
			 * PPI */
      central packvar,		/* central angle for (arcmin) for coord.
			 * transform */
      raster packvar,		/* raster number of this scan */
      init_angle packvar,		/* initial azim/elev of scan (arcmin) */
      angle_inc packvar,		/* angle increment between scans (arcmin)
				 * ST?? */
      gate_spacing packvar,	/* gate spacing in metres */
      range_offset packvar;	/* gate offset in metres, from A scope */
//   int
//      bytes_read;   /* Number of bytes read in current raster */
   float
      frequency,
      (*field_function[MAX_FIELDS]) (short);
}   RasterInfo;

typedef struct {
   uword
      ngates,avegates,averays,index,
      day,month,year, /*These change within a raster so can be useful */
      raw_day,raw_month,raw_year; /* of non-averaged rays */
   uchar
      nfields,	/* the number of raw fields stored in each ray */
      ntotal,	/* the total number of fields, both raw and derived */
      field_type[MAX_FIELDS], /* for both raw and derived fields */
      range_normalised,
      correct_range;
   float
      elevation,azimuth,      /* in degrees */
      utc,                     /* in decimal hours */
      raw_utc,                     /* in decimal hours */
      last_utc,       /* necessary for old cloud time stamping via az and el */
      sum_elevation,sum_azimuth,      /* in degrees */
      sum_utc,                     /* in decimal hours */
      v_floor,   /* rejection threshold for v in dB above the 
                                       rejection threshold for Z */
      sigma_v_floor,   /* rejection threshold for sigma_v in dB above the 
                                       rejection threshold for Z */
      nez;   /* Noise-equivalent reflectivity at 1 km; =0 if no processing */
   int
      field_info[MAX_FIELDS]; /* some of the derived fields need somewhere to 
            store the number of the parameter they are derived from (for instance) */
   long int
      maxsize;
   word
      hour_mod;     /* necessary for old cloud time stamping via az and el */
   float
      range_offset,   /* not the same as the RasterInformation version if gates are averaged */
      *range,   /* WITHOUT applying range correction */
      *data[MAX_FIELDS];
   double
      *sum[MAX_FIELDS];
} Ray;


typedef struct {
    short int
        azimuth,		/* azimuth of ray in arcminutes */
        elevation;		/* elevation of ray in arcminutes */
}   RayHeader;

typedef struct {
    unsigned short int
        mins_since_midnight,
        centisecs;         /* since last minute */
}   TimeStamp;





/* *** PROTOTYPES *** */
/* functions for looking at summary files */
int CHILGetFileList(uword day,uword month,uword year,int Argc,char **Argv,FileList * filelist);
FILE *CHILOpenSummary(char * datfilename);
/* functions used to open and close radar files: */
int CHILOpen(char *,ChilboltonFile *);
int CHILClose(ChilboltonFile *);
char *CHILSecondarySumDir(char * sec_sum_dir);
/* funtions used to negotiate radar files: */
int CHILReadInformationRecord(ChilboltonFile *,RasterInfo *);
int CHILNextRaster(ChilboltonFile *,RasterInfo *);
int CHILSeekRaster(ChilboltonFile *,uword required_raster);
int CHILSkipScans(ChilboltonFile *,RasterInfo *,int);
int CHILSkipRay(ChilboltonFile *,RasterInfo *);
/* ray functions */
int CHILInitRay(const RasterInfo *,int avegates, int averays,Ray *);
int CHILReInitRay(const RasterInfo *,int avegates, int averays,Ray *);
int CHILFreeRay(Ray *);
int CHILReadRawRay(ChilboltonFile *,RasterInfo *,Ray *);
int CHILReadRay(ChilboltonFile *,RasterInfo *,Ray *);
int CHILNextRay(ChilboltonFile *,const RasterInfo *);
int CHILSubtractNoise(Ray *ray,int noisegates,float factor,float nosigval);
int CHILSubtractOscillatorNoise(Ray *ray,int noisegates,float factor,float nosigval);
int CHILCleanRay(Ray *ray,float nosigval);
int CHILSuppressClutter(Ray *ray,float nosigval);
int CHILRangeCorrection(RasterInfo *,Ray *ray,float nosigval);
int CHILCalibrateZh(Ray *ray,float Zcal,float nosigval);
/* debugging functions */
int CHILPrintFileInformation(ChilboltonFile *);
int CHILPrintInformationRecord(RasterInfo *);
int CHILPrintRayInformation(Ray * ray);
/* information functions: */
char *CHILFieldDescription(const Ray *ray,int field,char * desc);
char *CHILFieldName(const Ray *ray,int field,char * name);
char *CHILFieldUnits(const Ray *ray,int field,char * units);
char *CHILSystemName(const RasterInfo *info,char *name);
/* conversion functions */
float CHILraw(short);
float CHILzh(short);
float CHILzh_highprecision(short);
float CHILzdr(short);
float CHILzdr_highprecision(short);
float CHILldr(short);
float CHILphidp_old(short);
float CHILphidp(short);
float CHILvel_old(short);
float CHILvel(short);
/* utilities */
void CHILIncrDate(uword *day,uword *month,uword *year);
void CHILDecrDate(uword *day,uword *month,uword *year);



